import sys
from colorama import init, Fore, Style
import signal
from zenif.schema import Schema, StringF
import shutil

init(autoreset=True)


class BasePrompt:
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        self.message = message
        self.schema = schema
        self.id = id
        if schema and not id:
            raise ValueError("You must have an ID in order to use a schema.")
        if schema and id:
            self.field = schema.fields.get(id)
            if not self.field:
                raise ValueError(f"Field '{id}' not found in the schema.")
        else:
            self.field = None

    def validate(self, value):
        try:
            if self.schema and self.id:
                is_valid, errors, _ = self.schema.validate({self.id: value})
                if not is_valid:
                    return errors.get(self.id, ["Invalid input"])[0].rstrip(".")
            elif self.field:
                self.field.validate(value)
            return None
        except ValueError as e:
            return str(e)

    @staticmethod
    def _get_key():
        def handle_interrupt(signum, frame):
            raise KeyboardInterrupt()

        if sys.platform.startswith("win"):
            import msvcrt

            # Set up the interrupt handler
            signal.signal(signal.SIGINT, handle_interrupt)

            try:
                while True:
                    if msvcrt.kbhit():
                        char = msvcrt.getch().decode("utf-8")
                        if char == "\x03":  # Ctrl+C
                            raise KeyboardInterrupt()
                        return char
            finally:
                # Reset the interrupt handler
                signal.signal(signal.SIGINT, signal.SIG_DFL)

        else:
            import termios
            import tty

            fd = sys.stdin.fileno()
            old_settings = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                # Set up the interrupt handler
                signal.signal(signal.SIGINT, handle_interrupt)

                while True:
                    char = sys.stdin.read(1)
                    if char == "\x03":  # Ctrl+C
                        raise KeyboardInterrupt()
                    if char == "\x1b":
                        # Handle escape sequences (e.g., arrow keys)
                        next_char = sys.stdin.read(1)
                        if next_char == "[":
                            last_char = sys.stdin.read(1)
                            return f"\x1b[{last_char}"
                    return char
            finally:
                # Reset terminal settings and interrupt handler
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                signal.signal(signal.SIGINT, signal.SIG_DFL)

    @staticmethod
    def _print_prompt(
        prompt: str = "",
        value: str = "",
        default: any = None,
        options: list[str] | None = None,
        default_option: str | None = None,
        error: str | None = None,
    ):
        sys.stdout.write(f"\033[2K\r{Fore.GREEN}? {Fore.CYAN}{prompt}{Fore.RESET}")
        if default and not options:
            sys.stdout.write(f" {Fore.CYAN}{Style.DIM}({default}){Style.RESET_ALL}")
        if options:
            if default_option:
                options[
                    [option.lower() for option in options].index(default_option.lower())
                ] = options[
                    [option.lower() for option in options].index(default_option.lower())
                ].upper()
            if len(options) == 2:
                sys.stdout.write(
                    f" {Fore.CYAN}{Style.DIM}[{options[0]}/{options[1]}]{Style.RESET_ALL}"
                )
            else:
                sys.stdout.write(
                    f" {Fore.CYAN}{Style.DIM}[{"".join(options)}]{Style.RESET_ALL}"
                )
        sys.stdout.write(f"{Fore.CYAN}: {Fore.YELLOW}{value}")
        if error:
            sys.stdout.write(f"  {Fore.RED}{error}\033[{2 + len(error)}D")
        sys.stdout.flush()


class TextPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self._default: str | None = None

    def default(self, value: str) -> "TextPrompt":
        self._default = value
        return self

    def ask(self) -> str:
        value = ""
        while True:
            error = self.validate(value or self._default or "")
            marker = "..."
            width = (
                shutil.get_terminal_size().columns
                - len(self.message)
                - len(error or "")
                - len(self._default or "")
                - (6 if self._default else 4)
                - (2 if error else 0)
            )
            truncated_value = (
                marker + value[-(width - len(marker)) :]
                if len(value) > width
                else value
            )

            self._print_prompt(
                self.message, truncated_value, self._default, error=error
            )
            char = self._get_key()
            if char == "\r":  # Enter key
                if not error and (value or self._default):
                    self._print_prompt(
                        self.message, value or self._default, self._default
                    )
                    print()
                    return value or self._default
            elif char == "\x7f":  # Backspace
                value = value[:-1]
            elif char not in ("\x1b[A", "\x1b[B", "\x1b[C", "\x1b[D"):
                value += char


class PasswordPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)

    def ask(self) -> str:
        value = ""
        while True:
            error = self.validate(value or "")
            masked_value = "*" * len(value)
            marker = "..."
            width = (
                shutil.get_terminal_size().columns
                - len(self.message)
                - len(error or "")
                - 2
                if error
                else 0 - 4
            )
            truncated_value = (
                marker + masked_value[-(width - len(marker)) :]
                if len(masked_value) > width
                else masked_value
            )
            self._print_prompt(self.message, truncated_value, error=error)
            char = self._get_key()
            if char == "\r":  # Enter key
                if not error and value:
                    print()  # Move to next line after input
                    return value
            elif char == "\x7f":  # Backspace
                value = value[:-1]
            elif char not in ("\x1b[A", "\x1b[B", "\x1b[C", "\x1b[D"):
                value += char


class ConfirmPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self._default: bool | None = None

    def default(self, value: bool) -> "ConfirmPrompt":
        self._default = value
        return self

    def ask(self) -> bool:
        options = (
            ["y", "N"]
            if self._default is False
            else ["Y", "n"] if self._default is True else ["y", "n"]
        )
        while True:
            self._print_prompt(
                self.message,
                options=options,
                default_option=(
                    "Y"
                    if self._default is True
                    else "N" if self._default is False else None
                ),
            )
            key = self._get_key().lower()
            result = (
                key == "y"
                if key in ("y", "n")
                else (
                    self._default if key == "\r" and self._default is not None else None
                )
            )
            if result is not None:
                error = self.validate(result or "")
                if not error:
                    self._print_prompt(
                        self.message,
                        value="Yes" if result else "No",
                        options=options,
                        default_option=(
                            "Y"
                            if self._default is True
                            else "N" if self._default is False else None
                        ),
                    )
                    print()
                    return result
                else:
                    self._print_prompt(self.message, error=error)


class ChoicePrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        choices: list[str],
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self.choices = choices

        # Check if the field is a StringF
        if not isinstance(self.field, StringF):
            field_type = type(self.field).__name__
            error_message = (
                f"ChoicePrompt requires a StringF field, but got {field_type}"
            )
            raise TypeError(error_message)

    def ask(self) -> str:
        current = 0
        print(
            f"{Fore.GREEN}? {Fore.CYAN}{self.message}:{Fore.RESET}\n{Style.DIM}  Use Up/Down to navigate and Enter to select"
        )
        while True:
            for i, choice in enumerate(self.choices):
                if i == current:
                    print(f"{Fore.YELLOW}{Style.NORMAL}> {choice}{Fore.RESET}")
                else:
                    print(f"{Fore.YELLOW}{Style.DIM}  {choice}{Fore.RESET}")

            key = self._get_key()
            if key == "\r":  # Enter key
                result = self.choices[current]
                error = self.validate(result or "")
                if not error:
                    for _ in range(len(self.choices) + 2):
                        print(f"\033[1A\033[2K", end="")
                    self._print_prompt(self.message, result)
                    print()  # Move to next line
                    return result
                else:
                    for _ in range(len(self.choices) + 2):
                        print(f"\033[1A\033[2K", end="")
                    self._print_prompt(self.message, error=error)
                    print()
                    print(
                        f"{Fore.GREEN}? {Fore.CYAN}{self.message}:{Fore.RESET}\n{Style.DIM}  Use Up/Down to navigate and Enter to select"
                    )
            elif key == "\x1b[A" and current > 0:  # Up arrow
                current -= 1
            elif key == "\x1b[B" and current < len(self.choices) - 1:  # Down arrow
                current += 1

            print(f"\033[{len(self.choices) + 1}A")  # Move cursor up to redraw choices


class CheckboxPrompt(BasePrompt):

    # TODO: Only show output on Enter like everything else

    def __init__(
        self,
        message: str,
        choices: list[str],
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self.choices = choices

    def ask(self) -> list[str]:
        selected = [False] * len(self.choices)
        current = 0
        print(
            f"{Fore.GREEN}? {Fore.CYAN}{self.message}:{Fore.RESET}\n{Style.DIM}  Use Up/Down to navigate, Space to select, and Enter to confirm"
        )
        i = True
        while True:
            for i, (choice, is_selected) in enumerate(zip(self.choices, selected)):
                if i == current:
                    print(f"{Fore.YELLOW}{Style.DIM}X{Style.NORMAL}", end="")
                else:
                    print(f"{Fore.YELLOW} ", end="")
                print(
                    f"\r{f"{Fore.YELLOW}{"\033[4m" if i == current else ""}X\033[0m" if is_selected else '\033[1C'} {Fore.YELLOW}{Style.DIM}{choice}{Fore.RESET}"
                )

            if i:
                i = False
                result = [
                    choice
                    for choice, is_selected in zip(self.choices, selected)
                    if is_selected
                ]
                error = self.validate(result)

                print(f"\033[{len(self.choices) + 2}A", end="")
                self._print_prompt(self.message, error=f"{error if error else ""}\n")
                print(
                    f"\r{Fore.RESET}{Style.DIM}  Use Up/Down to navigate, Space to select, and Enter to confirm\033[{len(self.choices)}B"
                )

            key = self._get_key()
            if key == " ":  # Space
                selected[current] = not selected[current]

            result = [
                choice
                for choice, is_selected in zip(self.choices, selected)
                if is_selected
            ]
            error = self.validate(result)

            print(f"\033[{len(self.choices) + 2}A", end="")
            self._print_prompt(self.message, error=f"{error if error else ""}\n")
            print(
                f"\r{Fore.RESET}{Style.DIM}  Use Up/Down to navigate, Space to select, and Enter to confirm\033[{len(self.choices)}B"
            )

            if key == "\r" and not error:
                return result
            elif key == "\x1b[A" and current > 0:  # Up arrow
                current -= 1
            elif key == "\x1b[B" and current < len(self.choices) - 1:  # Down arrow
                current += 1

            print(f"\033[{len(self.choices) + 1}A")  # Move cursor up to redraw choices


class NumberPrompt(BasePrompt):
    def __init__(
        self,
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ):
        super().__init__(message, schema, id)
        self._default: int | None = None
        self._commas: bool = False

    def default(self, value: int) -> "NumberPrompt":
        self._default = value
        return self

    def commas(self) -> "NumberPrompt":
        self._commas = True
        return self

    def ask(self) -> int:
        value = ""
        while True:
            try:
                int_value = (
                    int(value) if value else self._default if self._default else 0
                )
                error = self.validate(int_value)
            except ValueError:
                error = "Please enter a valid number."

            marker = "..."
            width = (
                shutil.get_terminal_size().columns
                - len(self.message)
                - len(error or "")
                - len(str(self._default or ""))
                - (7 if self._default else 5)
                - (2 if error else 0)
            )
            formatted_value = f"{int_value:,}" if self._commas and value else str(value)
            truncated_value = (
                marker + formatted_value[-(width - len(marker)) :]
                if len(formatted_value) > width
                else formatted_value
            )

            self._print_prompt(
                self.message, truncated_value, self._default, error=error
            )
            char = self._get_key()
            if char == "\r":  # Enter key
                if not error and (value or self._default is not None):
                    print()  # Move to next line after input
                    return float(value) if value else self._default
            elif char == "\x7f":  # Backspace
                value = value[:-1]
            elif char.isdigit() or char in ("-"):
                value += char


class Prompt:
    @staticmethod
    def text(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> TextPrompt:
        return TextPrompt(message, schema, id)

    @staticmethod
    def password(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> PasswordPrompt:
        return PasswordPrompt(message, schema, id)

    @staticmethod
    def confirm(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> ConfirmPrompt:
        return ConfirmPrompt(message, schema, id)

    @staticmethod
    def choice(
        message: str,
        choices: list[str],
        schema: Schema | None = None,
        id: str | None = None,
    ) -> ChoicePrompt:
        return ChoicePrompt(message, choices, schema, id)

    @staticmethod
    def checkbox(
        message: str,
        choices: list[str],
        schema: Schema | None = None,
        id: str | None = None,
    ) -> CheckboxPrompt:
        return CheckboxPrompt(message, choices, schema, id)

    @staticmethod
    def number(
        message: str,
        schema: Schema | None = None,
        id: str | None = None,
    ) -> NumberPrompt:
        return NumberPrompt(message, schema, id)



from zenif.cli import CLI, arg, kwarg, Prompt, install_setup_command
from zenif.schema import Schema, StringF, IntegerF, ListF, Length, Value, EmailValidator
import os

cli = CLI()

install_setup_command(cli, os.path.abspath(__file__))


@cli.command
@arg("name", help="Name to greet")
@kwarg("--greeting", default="Hello", help="Greeting to use")
@kwarg("--shout", is_flag=True, help="Print in uppercase")
def greet(name: str, greeting: str, shout: bool = False):
    """Greet a person."""
    message = f"{greeting}, {name}!"
    if shout:
        message = message.upper()
    return message


@cli.command
def test_prompts():
    """Test all available prompts"""

    class OddOrEven:
        def __init__(self, parity: str = "even"):
            self.parity = 1 if parity == "odd" else 0

        def __call__(self, value):
            if value % 2 != self.parity:
                raise ValueError(
                    f"Must be an {'even' if self.parity == 0 else 'odd'} number."
                )

    schema = Schema(
        name=StringF().name("name").has(Length(min=3, max=50)),
        age=IntegerF()
        .name("age")
        .has(Value(min=18, max=120))
        .has(OddOrEven(parity="odd")),
        interests=ListF().name("interests").item_type(StringF()).has(Length(min=3)),
        fav_interest=StringF().name("fav_interest"),
        email=StringF().name("email").has(EmailValidator()),
    ).all_optional()

    name = Prompt.text("Enter your name", schema=schema, id="name").ask()
    age = Prompt.number("Enter your age", schema=schema, id="age").ask()
    interests = Prompt.checkbox(
        "Select your interests",
        choices=["Reading", "Gaming", "Sports", "Cooking", "Travel"],
        schema=schema,
        id="interests",
    ).ask()
    fav_interest = Prompt.choice(
        "Select your favorite interest",
        choices=interests,
        schema=schema,
        id="fav_interest",
    ).ask()
    email = Prompt.text("Enter your email", schema=schema, id="email").ask()

    print(f"Name: {name}")
    print(f"Age: {age}")
    print(f"Interests: {', '.join(interests)}")
    print(f"Favorite Interest: {fav_interest}")
    print(f"Email: {email}")


if __name__ == "__main__":
    cli.run()



import time
from functools import wraps
import signal
from collections import deque
from typing import Callable, TypeVar, Union
from inspect import signature
from threading import Thread
from collections import OrderedDict
import time
import functools
import cProfile
import pstats
import io
import tracemalloc
from .exceptions import TimeoutError, RateLimitError
from ..log import Logger

logger = Logger(ruleset={"timestamps": {"always_show": True}})

T = TypeVar("T")


def retry(
    max_retries: int = 3, delay: float = 1.0
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def decorator_retry(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper_retry(*args: any, **kwargs: any) -> T:
            attempts = 0
            while attempts < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts >= max_retries:
                        raise
                    time.sleep(max(0, delay))  # Ensure non-negative delay
            raise RuntimeError("Exceeded maximum retries")

        return wrapper_retry

    return decorator_retry


def retry_expo(
    max_retries: int = 3, initial_delay: float = 1.0
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def decorator_retry_expo_backoff(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper_retry_expo_backoff(*args: any, **kwargs: any) -> T:
            attempts = 0
            while attempts < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempts += 1
                    if attempts >= max_retries:
                        raise
                    delay = initial_delay * (2 ** (attempts - 1))
                    time.sleep(max(0, delay))  # Ensure non-negative delay
            raise RuntimeError("Exceeded maximum retries")

        return wrapper_retry_expo_backoff

    return decorator_retry_expo_backoff


def timeout(seconds: float) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def decorator_timeout(func: Callable[..., T]) -> Callable[..., T]:
        def _handle_timeout(signum: int, frame: Union[any, None]) -> None:
            raise TimeoutError(
                f"Function {func.__name__} timed out after {seconds} seconds"
            )

        wraps(func)

        def wrapper_timeout(*args: any, **kwargs: any) -> T:
            # Use SIGALRM only on Unix-like systems
            if hasattr(signal, "SIGALRM"):
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(int(seconds))
                try:
                    return func(*args, **kwargs)
                finally:
                    signal.alarm(0)  # Disable the alarm
            else:
                # Fallback for systems without SIGALRM (e.g., Windows)
                return func(*args, **kwargs)

        return wrapper_timeout

    return decorator_timeout


def rate_limiter(
    calls: int, period: float, immediate_fail: bool = True
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def decorator_rate_limiter(func: Callable[..., T]) -> Callable[..., T]:
        call_times: deque = deque(maxlen=calls)

        @wraps(func)
        def wrapper_rate_limiter(*args: any, **kwargs: any) -> T:
            current_time = time.monotonic()  # Use monotonic time for better precision
            while call_times and current_time - call_times[0] > period:
                call_times.popleft()

            if len(call_times) < calls:
                call_times.append(current_time)
                return func(*args, **kwargs)
            else:
                if immediate_fail:
                    raise RateLimitError(
                        f"Function {func.__name__} rate-limited. Try again later."
                    )
                else:
                    wait_time = max(0, period - (current_time - call_times[0]))
                    time.sleep(wait_time)
                    return wrapper_rate_limiter(*args, **kwargs)

        return wrapper_rate_limiter

    return decorator_rate_limiter


def trace(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def wrapper_trace(*args: any, **kwargs: any) -> T:
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        print(f"Calling {func.__name__}({signature})")
        try:
            result = func(*args, **kwargs)
            print(f"{func.__name__} returned {result!r}")
            return result
        except Exception as e:
            print(f"{func.__name__} raised {type(e).__name__}: {str(e)}")
            raise

    return wrapper_trace


def suppress_exceptions(func: Callable[..., T | None]) -> Callable[..., T | None]:
    @wraps(func)
    def wrapper_suppress_exceptions(*args: any, **kwargs: any) -> T | None:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(
                f"Exception suppressed in {func.__name__}: {type(e).__name__}: {str(e)}"
            )
            return None

    return wrapper_suppress_exceptions


def deprecated(
    func: Callable[..., T] | None = None, *, expected_removal: str | None = None
) -> Callable[..., T]:
    def decorator(f: Callable[..., T]) -> Callable[..., T]:
        @wraps(f)
        def wrapper_deprecated(*args: any, **kwargs: any) -> T:
            warning_message = (
                f"{f.__name__} is deprecated and will be removed in a future version"
            )
            if expected_removal:
                warning_message += f" (expected removal: {expected_removal})"
            logger.warning(warning_message)
            return f(*args, **kwargs)

        return wrapper_deprecated

    if func is None:
        return decorator
    else:
        return decorator(func)


def type_check(
    arg_types: tuple[type, ...] | None = None, return_type: type | None = None
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    def decorator_type_check(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper_type_check(*args: any, **kwargs: any) -> T:
            if arg_types:
                for arg, expected_type in zip(args, arg_types):
                    if not isinstance(arg, expected_type):
                        raise TypeError(
                            f"Argument {arg!r} does not match type {expected_type}"
                        )
            result = func(*args, **kwargs)
            if return_type and not isinstance(result, return_type):
                raise TypeError(
                    f"Return value {result!r} does not match type {return_type}"
                )
            return result

        return wrapper_type_check

    return decorator_type_check


def log_execution_time(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def wrapper_log_execution_time(*args: any, **kwargs: any) -> T:
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        print(
            f"Execution time for {func.__name__}: {end_time - start_time:.6f} seconds"
        )
        return result

    return wrapper_log_execution_time


def cache(
    func: Callable[..., T] | None = None, *, max_size: int | None = None
) -> Callable[..., T]:
    def decorator(f: Callable[..., T]) -> Callable[..., T]:
        cache_dict: OrderedDict = OrderedDict()

        @wraps(f)
        def wrapper_cache(*args: any, **kwargs: any) -> T:
            key = (*args, *sorted(kwargs.items()))
            if key not in cache_dict:
                result = f(*args, **kwargs)
                cache_dict[key] = result
                if max_size is not None and len(cache_dict) > max_size:
                    cache_dict.popitem(last=False)  # Remove the oldest item
            return cache_dict[key]

        # Add a method to clear the cache
        wrapper_cache.clear_cache = cache_dict.clear

        return wrapper_cache

    if func is None:
        return decorator
    else:
        return decorator(func)


def singleton(cls):
    instances = {}

    @wraps(cls)
    def get_instance(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return get_instance


def enforce_types(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        sig = signature(func)
        bound_args = sig.bind(*args, **kwargs)
        for name, value in bound_args.arguments.items():
            if name in sig.parameters:
                param = sig.parameters[name]
                if param.annotation != param.empty:
                    if not isinstance(value, param.annotation):
                        raise TypeError(f"Argument {name} must be {param.annotation}")

        result = func(*args, **kwargs)
        if sig.return_annotation != sig.empty:
            if not isinstance(result, sig.return_annotation):
                raise TypeError(f"Return value must be {sig.return_annotation}")
        return result

    return wrapper


def retry_on_exception(exceptions, max_retries=3, delay=1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    if attempt == max_retries - 1:
                        raise
                    time.sleep(delay)
            return func(*args, **kwargs)  # This line should never be reached

        return wrapper

    return decorator


def background_task(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        thread = Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper


_profiler_active = False


def profile(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _profiler_active

        if not _profiler_active:
            _profiler_active = True
            profiler = cProfile.Profile()
            try:
                tracemalloc.start()
                start_time = time.time()
                profiler.enable()
                result = func(*args, **kwargs)
                profiler.disable()
                end_time = time.time()
            finally:
                try:
                    current, peak = tracemalloc.get_traced_memory()
                except RuntimeError:
                    current, peak = 0, 0
                tracemalloc.stop()
                _profiler_active = False

            s = io.StringIO()
            ps = pstats.Stats(profiler, stream=s).sort_stats("cumulative")
            ps.print_stats(10)  # Print top 10 lines

            print(f"Function: {func.__name__}")
            print(f"{"Time taken:".ljust(21)} {end_time - start_time:.4f} seconds")
            print(f"Current memory usage: {current / 10**6:.6f} MB")
            print(f"{"Peak memory usage:".ljust(21)} {peak / 10**6:.6f} MB")
            print("Profile:")
            print(s.getvalue())
        else:
            result = func(*args, **kwargs)

        return result

    return wrapper



from .core import CLI
from .decorators import arg, kwarg
from .prompt import Prompt
from .install_command import install_setup_command


__all__ = [
    # CLI()
    "CLI",
    # CLI().command decorators
    "arg",
    "kwarg",
    # interactive prompt utils
    "Prompt",
    # setup command installer
    "install_setup_command",
    # Schema
    "Schema",
    "ValidationResult",
    # Fields
    "Field",  # base class
    "String",
    "Integer",
    "Float",
    "Boolean",
    "List",
    "Dict",
    "Email",
    # Validators
    "Validator",  # base class
    "Length",
    "Value",
]



from typing import Callable


def arg(*args, **kwargs):
    """Decorator to add an argument to a CLI command."""

    def decorator(func: Callable) -> Callable:
        if not hasattr(func, "_arguments"):
            func._arguments = []
        func._arguments.append((args, kwargs))
        return func

    return decorator


def kwarg(*args, **kwargs):
    """Decorator to add a keyword argument to a CLI command."""
    kwargs["is_option"] = True
    return arg(*args, **kwargs)



from datetime import datetime, UTC
from shutil import get_terminal_size as tsize
from threading import current_thread
from random import choice
from black import format_str, Mode
from io import StringIO
from re import sub
from copy import deepcopy

from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import Terminal256Formatter as tformatter
from pygments.styles import get_style_by_name

from .utils import strip_ansi, strip_unsafe_objs, strip_repr_id, wrap
from .template import TemplateEngine


class Ruleset:

    def __init__(self, rules: dict[str, any], defaults: dict[str, any]):
        self._rules = self._merge_with_defaults(deepcopy(rules), deepcopy(defaults))
        self.dict = self.Dict(self._rules)
        for category, settings in self._rules.items():
            setattr(
                self,
                category,
                type(
                    f"{category.title().replace('_', '')}Rules", (), deepcopy(settings)
                )(),
            )

    def _merge_with_defaults(self, rules: dict[str, any], defaults: dict[str, any]):
        merged = defaults.copy()
        for category, settings in rules.items():
            if category in merged:
                merged[category].update(settings)
            else:
                merged[category] = settings
        return merged

    class Dict:
        def __init__(self, rules):
            self.all = rules
            for category, settings in rules.items():
                setattr(self, category, deepcopy(settings))


class BaseHandler:

    def __init__(self, defaults: dict[str, any]):
        self.output_streams = []
        self.stream_rulesets = {}
        self.template_engine = TemplateEngine()
        self.previous_timestamp = None
        self.defaults = deepcopy(defaults)

    def add(self, stream: str | object, ruleset: dict[str, any] = None):
        if stream not in self.output_streams:
            self.output_streams.append(stream)
            if ruleset:
                self.stream_rulesets[stream] = Ruleset(deepcopy(ruleset), self.defaults)
            else:
                self.stream_rulesets[stream] = Ruleset({}, self.defaults)
        return stream

    def remove(self, stream: str | object):
        if stream in self.output_streams:
            self.output_streams.remove(stream)
            if stream in self.stream_rulesets:
                del self.stream_rulesets[stream]

    def modify(
        self, stream: str | object, ruleset: dict[str, any], use_original: bool = False
    ):
        if stream not in self.stream_rulesets:
            raise ValueError(f"Stream {stream} not found in handler.")

        if use_original:
            new_rules = deepcopy(self.defaults)
            new_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(new_rules, self.defaults)
        else:
            current_rules = deepcopy(self.stream_rulesets[stream]._rules)
            current_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(current_rules, self.defaults)

    def process_message(
        self,
        message: str,
        level: dict[str, str | int],
        metadata: dict[str, any],
        ruleset: dict[str, any],
    ):
        terminal_width = tsize().columns

        context = {
            "timestamp": self._define_timestamp(ruleset),
            "filename": metadata.get("file_name", ""),
            "wrapfunc": metadata.get("wrapping_func", ""),
            "linenum": metadata.get("line_number", ""),
            "level": level["name"],
        }

        log_line = self.template_engine.process(
            ruleset.log_line.format, context, level["name"]
        )
        log_line = sub(r"\033\[(\d+)C", lambda m: " " * int(m.group(1)), log_line)
        log_line = log_line if ruleset.formatting.ansi else strip_ansi(log_line)

        message_space = terminal_width - self.template_engine.processed.length
        message_indent = self.template_engine.processed.length

        repr_id = "".join(
            choice("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")
            for _ in range(4)
        )

        formatted_message = self._format_message(
            message, ruleset, repr_id, message_space
        )
        log_metadata = self._generate_metadata(
            metadata, ruleset, message_space, terminal_width
        )

        formatted_message = (
            formatted_message
            if ruleset.formatting.ansi
            else strip_ansi(formatted_message)
        )

        log_line += log_metadata

        lines = wrap(formatted_message, message_space)

        log_output = StringIO()

        log_output.write(
            f"{log_line}{''.join([f'{line}\n' if i == 0 else f'\033[{message_indent}C{line}\033[0m\n' for i, line in enumerate(lines)])}"
        )

        return log_output.getvalue()

    def _define_timestamp(self, ruleset: dict[str, any]):
        now = datetime.now(UTC if ruleset.timestamps.use_utc else None)
        timestamp = now.strftime("%H:%M:%S")
        if timestamp == self.previous_timestamp and not ruleset.timestamps.always_show:
            return f"\033[{len(timestamp)}C"
        self.previous_timestamp = timestamp
        return timestamp

    def _format_message(
        self, message: str, ruleset: dict[str, any], repr_id: str, message_space: int
    ):
        if ruleset.formatting.pretty_print and isinstance(message, (list, dict, tuple)):
            formatted = highlight(
                strip_repr_id(
                    format_str(
                        str(strip_unsafe_objs(message, repr_id)),
                        mode=Mode(
                            line_length=(
                                ruleset.formatting.fixed_format_width
                                if ruleset.formatting.fixed_format_width > 0
                                else message_space
                            )
                        ),
                    ),
                    repr_id,
                ),
                PythonLexer(),
                tformatter(style=get_style_by_name("one-dark")),
            )
            return formatted
        return str(message)

    def _generate_metadata(
        self,
        metadata: dict[str, any],
        ruleset: dict[str, any],
        message_space: int,
        terminal_width: int,
    ):
        if not ruleset.metadata.show_metadata:
            return ""

        metadata_items = []
        if ruleset.metadata.include_timestamp:
            metadata_items.append(f"[tms: {self._define_timestamp(ruleset)}]")
        if ruleset.metadata.include_level_name:
            metadata_items.append(f"[lvl: {metadata['level']}]")
        if ruleset.metadata.include_thread_name:
            metadata_items.append(f"[thr: {current_thread().name}]")
        if ruleset.metadata.include_file_name:
            metadata_items.append(f"[fl: {metadata['file_name']}]")
        if ruleset.metadata.include_wrapping_function:
            metadata_items.append(f"[wfc: {metadata['wrapping_func']}]")
        if ruleset.metadata.include_function:
            metadata_items.append(f"[fnc: {metadata['function']}]")
        if ruleset.metadata.include_line_number:
            metadata_items.append(f"[ln: {metadata['line_number']}]")
        if ruleset.metadata.include_value_count:
            metadata_items.append(f"[vlc: {metadata['value_count']}]")

        metadata_str = " ".join(metadata_items)
        metadata_space = terminal_width - message_space - len(strip_ansi(metadata_str))
        return f"\033[0m\033[2m{metadata_str.rjust(metadata_space)}\033[0m"


class StreamHandler(BaseHandler):

    def __init__(self, defaults: dict[str, any]):
        super().__init__(defaults)
        self.output_streams = []
        self.stream_rulesets = {}

    def add(self, stream: object, ruleset: dict[str, any] = None):
        if stream not in self.output_streams:
            self.output_streams.append(stream)
            if ruleset:
                self.stream_rulesets[stream] = Ruleset(deepcopy(ruleset), self.defaults)
            else:
                self.stream_rulesets[stream] = Ruleset({}, self.defaults)
        return stream

    def modify(
        self, stream: object, ruleset: dict[str, any], use_original: bool = False
    ):
        if stream not in self.stream_rulesets:
            raise ValueError(f"Stream {stream} not found in handler.")

        if use_original:
            new_rules = deepcopy(self.defaults)
            new_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(new_rules, self.defaults)
        else:
            current_rules = deepcopy(self.stream_rulesets[stream].dict.all)
            current_rules.update(deepcopy(ruleset))
            self.stream_rulesets[stream] = Ruleset(current_rules, self.defaults)

    def write(
        self, message: str, level_dict: dict[str, str | int], metadata: dict[str, any]
    ):
        for stream in self.output_streams:
            ruleset = self.stream_rulesets.get(stream)
            if ruleset:
                if not self._should_log(message, level_dict["level"], ruleset):
                    continue
                processed_message = self.process_message(
                    message, level_dict, metadata, ruleset
                )
                stream.write(processed_message)
            else:
                stream.write(message)
            stream.flush()

    def _should_log(self, message: str, level: int, ruleset: dict[str, any]):
        if level < ruleset.filtering.min_level:
            return False
        if any(
            substring in message for substring in ruleset.filtering.exclude_messages
        ):
            return False
        if ruleset.filtering.include_only_messages and not any(
            substring in message
            for substring in ruleset.filtering.include_only_messages
        ):
            return False
        return True


class FileHandler(BaseHandler):
    def __init__(self, defaults: dict[str, any]):
        super().__init__(defaults)
        self.file_streams = {}
        self.file_rulesets = {}

    def add(
        self, file_path: str, ruleset: dict[str, any] | None = None, reset: bool = False
    ):
        if file_path not in self.file_streams or reset:
            mode = "w" if reset else "a"
            self.file_streams[file_path] = open(file_path, mode)
            if ruleset:
                self.file_rulesets[file_path] = Ruleset(
                    deepcopy(ruleset), self.defaults
                )
            else:
                self.file_rulesets[file_path] = Ruleset({}, self.defaults)
        return file_path

    def remove(self, file_path: str):
        if file_path in self.file_streams:
            self.file_streams[file_path].close()
            del self.file_streams[file_path]
            if file_path in self.file_rulesets:
                del self.file_rulesets[file_path]

    def modify(
        self, file_path: str, ruleset: dict[str, any], use_original: bool = False
    ):
        if file_path in self.file_rulesets:
            if use_original:
                new_rules = deepcopy(self.defaults)
                new_rules.update(deepcopy(ruleset))
                self.file_rulesets[file_path] = Ruleset(new_rules, self.defaults)
            else:
                current_rules = deepcopy(self.file_rulesets[file_path]._rules)
                current_rules.update(deepcopy(ruleset))
                self.file_rulesets[file_path] = Ruleset(current_rules, self.defaults)

    def write(
        self, message: str, level_dict: dict[str, str | int], metadata: dict[str, any]
    ):
        for file_path, file_stream in self.file_streams.items():
            ruleset = self.file_rulesets.get(file_path)
            if ruleset:
                if not self._should_log(message, level_dict["level"], ruleset):
                    continue
                processed_message = self.process_message(
                    message, level_dict, metadata, ruleset
                )
                file_stream.write(strip_ansi(processed_message))
            else:
                file_stream.write(strip_ansi(message))
            file_stream.flush()

    def _should_log(self, message: str, level: int, ruleset: dict[str, any]):
        if level < ruleset.filtering.min_level:
            return False
        if any(
            substring in message for substring in ruleset.filtering.exclude_messages
        ):
            return False
        if ruleset.filtering.include_only_messages and not any(
            substring in message
            for substring in ruleset.filtering.include_only_messages
        ):
            return False
        return True

    def reset(self, file_path: str):
        if file_path in self.file_streams:
            self.remove(file_path)
            self.add(file_path, reset=True)


class Streams:
    def __init__(self, defaults: dict[str, any]):
        self.file = FileHandler(defaults)
        self.normal = StreamHandler(defaults)


class FHGroup:
    def __init__(self, file_handler: FileHandler, *items):
        self.file_handler = file_handler
        self.file_paths = []
        self.add(*items)

    def unwrap_fhgroups(self, items):
        new = []
        for item in items:
            if isinstance(item, str):
                new.append(item)
            elif isinstance(item, FHGroup):
                new.extend(self.unwrap_fhgroups(item.file_paths))
        return new

    def add(self, *items):
        items = self.unwrap_fhgroups(items)
        for item in items:
            if item not in self.file_paths:
                self.file_paths.append(item)
                self.file_handler.add(item)
        return [*items]

    def remove(self, *items: str):
        items = self.unwrap_fhgroups(items)
        for file_path in items:
            if file_path in self.file_paths:
                self.file_paths.remove(file_path)
                self.file_handler.remove(file_path)
        return [*items]

    def reset(self):
        for file_path in self.file_paths:
            self.file_handler.remove(file_path)
            self.file_handler.add(file_path, reset=True)

    def modify(self, ruleset: dict[str, any], use_original: bool = False):
        for file_path in self.file_paths:
            self.file_handler.modify(file_path, ruleset, use_original)

    def remove_all(self):
        for file_path in self.file_paths:
            self.file_handler.remove(file_path)
        self.file_paths.clear()


class SHGroup:

    def __init__(self, stream_handler: StreamHandler, *items):
        self.stream_handler = stream_handler
        self.streams = []
        self.add(*items)

    def unwrap_shgroups(self, items):
        new = []
        for item in items:
            if isinstance(item, SHGroup):
                new.extend(item.streams)
            else:
                new.append(item)
        return new

    def add(self, *items):
        items = self.unwrap_shgroups(items)
        for item in items:
            if item not in self.streams:
                self.streams.append(item)
                self.stream_handler.add(item)
        return [*items]

    def remove(self, *items: str):
        items = self.unwrap_shgroups(items)
        for item in items:
            if item in self.streams:
                self.streams.remove(item)
                self.stream_handler.remove(item)
        return [*items]

    def modify(self, ruleset: dict[str, any], use_original: bool = False):
        for stream in self.streams:
            self.stream_handler.modify(stream, ruleset, use_original)

    def remove_all(self):
        for stream in self.streams:
            self.stream_handler.remove(stream)
        self.streams.clear()
